#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from tests.fixtures import TempStoreClient

import pytest  # isort:skip

starlette = pytest.importorskip("starlette")  # isort:skip

import os

import mock
import urllib3
import wrapt
from starlette.applications import Starlette
from starlette.responses import PlainTextResponse
from starlette.staticfiles import StaticFiles
from starlette.testclient import TestClient

import elasticapm
from elasticapm import async_capture_span
from elasticapm.conf import constants
from elasticapm.contrib.starlette import ElasticAPM, make_apm_client
from elasticapm.utils.disttracing import TraceParent

pytestmark = [pytest.mark.starlette]

starlette_version_tuple = tuple(map(int, starlette.__version__.split(".")[:3]))

file_path, file_name = os.path.split(__file__)


@pytest.fixture
def app(elasticapm_client):
    app = Starlette()
    sub = Starlette()
    subsub = Starlette()
    app.mount("/sub", sub)
    sub.mount("/subsub", subsub)

    @app.exception_handler(Exception)
    async def handle_exception(request, exc):
        transaction_id = elasticapm.get_transaction_id()
        exc.transaction_id = transaction_id
        return PlainTextResponse(f"{transaction_id}", status_code=500)

    @app.route("/", methods=["GET", "POST"])
    async def hi(request):
        body = await request.body()
        with async_capture_span("test"):
            pass
        return PlainTextResponse(str(len(body)))

    @app.route("/hi/{name}", methods=["GET"])
    async def hi_name(request):
        name = request.path_params["name"]
        return PlainTextResponse("Hi {}".format(name))

    @app.route("/hello", methods=["GET", "POST"])
    async def hello(request):
        with async_capture_span("test"):
            pass
        return PlainTextResponse("ok")

    @app.route("/raise-exception", methods=["GET", "POST"])
    async def raise_exception(request):
        await request.body()
        raise ValueError()

    @app.route("/raise-base-exception", methods=["GET", "POST"])
    async def raise_base_exception(request):
        await request.body()
        raise Exception()

    @app.route("/hi/{name}/with/slash/", methods=["GET", "POST"])
    async def with_slash(request):
        return PlainTextResponse("Hi {}".format(request.path_params["name"]))

    @app.route("/hi/{name}/without/slash", methods=["GET", "POST"])
    async def without_slash(request):
        return PlainTextResponse("Hi {}".format(request.path_params["name"]))

    @app.route("/500/", methods=["GET"])
    async def with_500_status_code(request):
        return PlainTextResponse("Oops", status_code=500)

    @sub.route("/hi")
    async def hi_from_sub(request):
        return PlainTextResponse("sub")

    @subsub.route("/hihi/{name}")
    async def hi_from_sub(request):
        return PlainTextResponse(request.path_params["name"])

    @app.websocket_route("/ws")
    async def ws(websocket):
        await websocket.accept()
        await websocket.send_text("Hello, world!")
        await websocket.close()

    app.add_middleware(ElasticAPM, client=elasticapm_client)

    yield app

    elasticapm.uninstrument()


def test_get(app, elasticapm_client):
    client = TestClient(app)

    response = client.get(
        "/",
        headers={
            constants.TRACEPARENT_HEADER_NAME: "00-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-03",
            constants.TRACESTATE_HEADER_NAME: "foo=bar,bar=baz",
            "REMOTE_ADDR": "127.0.0.1",
        },
    )

    assert response.status_code == 200

    assert len(elasticapm_client.events[constants.TRANSACTION]) == 1
    transaction = elasticapm_client.events[constants.TRANSACTION][0]
    spans = elasticapm_client.spans_for_transaction(transaction)
    assert len(spans) == 1
    span = spans[0]

    assert transaction["name"] == "GET /"
    assert transaction["result"] == "HTTP 2xx"
    assert transaction["outcome"] == "success"
    assert transaction["type"] == "request"
    assert transaction["span_count"]["started"] == 1
    request = transaction["context"]["request"]
    assert request["method"] == "GET"
    assert request["socket"] == {"remote_address": "127.0.0.1"}

    response = transaction["context"]["response"]
    assert response["status_code"] == 200
    assert response["headers"]["content-type"] == "text/plain; charset=utf-8"

    assert span["name"] == "test"


@pytest.mark.parametrize("elasticapm_client", [{"capture_body": "all"}], indirect=True)
def test_post(app, elasticapm_client):
    client = TestClient(app)

    response = client.post(
        "/",
        headers={
            constants.TRACEPARENT_HEADER_NAME: "00-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-03",
            constants.TRACESTATE_HEADER_NAME: "foo=bar,bar=baz",
            "REMOTE_ADDR": "127.0.0.1",
        },
        data={"foo": "bar"},
    )

    assert response.status_code == 200

    assert len(elasticapm_client.events[constants.TRANSACTION]) == 1
    transaction = elasticapm_client.events[constants.TRANSACTION][0]
    spans = elasticapm_client.spans_for_transaction(transaction)
    assert len(spans) == 1
    span = spans[0]

    assert transaction["name"] == "POST /"
    assert transaction["result"] == "HTTP 2xx"
    assert transaction["outcome"] == "success"
    assert transaction["type"] == "request"
    assert transaction["span_count"]["started"] == 1
    request = transaction["context"]["request"]
    assert request["method"] == "POST"
    assert request["socket"] == {"remote_address": "127.0.0.1"}
    assert request["body"] == "foo=bar"

    assert span["name"] == "test"


def test_exception(app, elasticapm_client):
    client = TestClient(app)

    with pytest.raises(ValueError):
        client.get(
            "/raise-exception",
            headers={
                constants.TRACEPARENT_HEADER_NAME: "00-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-03",
                constants.TRACESTATE_HEADER_NAME: "foo=bar,bar=baz",
                "REMOTE_ADDR": "127.0.0.1",
            },
        )

    assert len(elasticapm_client.events[constants.TRANSACTION]) == 1
    transaction = elasticapm_client.events[constants.TRANSACTION][0]
    spans = elasticapm_client.spans_for_transaction(transaction)
    assert len(spans) == 0

    assert transaction["name"] == "GET /raise-exception"
    assert transaction["result"] == "HTTP 5xx"
    assert transaction["outcome"] == "failure"
    assert transaction["type"] == "request"
    request = transaction["context"]["request"]
    assert request["method"] == "GET"
    assert request["socket"] == {"remote_address": "127.0.0.1"}
    assert transaction["context"]["response"]["status_code"] == 500

    assert len(elasticapm_client.events[constants.ERROR]) == 1
    error = elasticapm_client.events[constants.ERROR][0]
    assert error["transaction_id"] == transaction["id"]
    assert error["exception"]["type"] == "ValueError"
    assert error["context"]["request"] == transaction["context"]["request"]


def test_failure_outcome_with_500_status_code(app, elasticapm_client):
    client = TestClient(app)

    client.get("/500/")

    assert len(elasticapm_client.events[constants.TRANSACTION]) == 1
    transaction = elasticapm_client.events[constants.TRANSACTION][0]
    spans = elasticapm_client.spans_for_transaction(transaction)
    assert len(spans) == 0

    assert transaction["name"] == "GET /500/"
    assert transaction["result"] == "HTTP 5xx"
    assert transaction["outcome"] == "failure"
    assert transaction["type"] == "request"
    request = transaction["context"]["request"]
    assert request["method"] == "GET"
    assert transaction["context"]["response"]["status_code"] == 500

    assert len(elasticapm_client.events[constants.ERROR]) == 0


@pytest.mark.parametrize("header_name", [constants.TRACEPARENT_HEADER_NAME, constants.TRACEPARENT_LEGACY_HEADER_NAME])
def test_traceparent_handling(app, elasticapm_client, header_name):
    client = TestClient(app)
    with mock.patch(
        "elasticapm.contrib.starlette.TraceParent.from_string", wraps=TraceParent.from_string
    ) as wrapped_from_string:
        response = client.get(
            "/",
            headers={
                header_name: "00-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-03",
                constants.TRACESTATE_HEADER_NAME: "foo=bar,baz=bazzinga",
            },
        )

    assert response.status_code == 200

    transaction = elasticapm_client.events[constants.TRANSACTION][0]

    assert transaction["trace_id"] == "0af7651916cd43dd8448eb211c80319c"
    assert transaction["parent_id"] == "b7ad6b7169203331"
    assert "foo=bar,baz=bazzinga" in wrapped_from_string.call_args[0]


def test_capture_headers_body_is_dynamic(app, elasticapm_client):
    client = TestClient(app)

    for i, val in enumerate((True, False)):
        elasticapm_client.config.update(str(i), capture_body="transaction" if val else "none", capture_headers=val)
        try:
            client.post("/", content="somedata", headers={"foo": "bar"})
        except TypeError:  # starlette < 0.21.0 used requests as base for TestClient, with a different API
            client.post("/", "somedata", headers={"foo": "bar"})

        elasticapm_client.config.update(str(i) + str(i), capture_body="error" if val else "none", capture_headers=val)
        with pytest.raises(ValueError):
            try:
                client.post("/raise-exception", content="somedata", headers={"foo": "bar"})
            except TypeError:
                client.post("/raise-exception", "somedata", headers={"foo": "bar"})

    assert "headers" in elasticapm_client.events[constants.TRANSACTION][0]["context"]["request"]
    assert "headers" in elasticapm_client.events[constants.TRANSACTION][0]["context"]["response"]
    assert elasticapm_client.events[constants.TRANSACTION][0]["context"]["request"]["body"] == "somedata"
    assert "headers" in elasticapm_client.events[constants.ERROR][0]["context"]["request"]
    assert elasticapm_client.events[constants.ERROR][0]["context"]["request"]["body"] == "somedata"

    assert "headers" not in elasticapm_client.events[constants.TRANSACTION][2]["context"]["request"]
    assert "headers" not in elasticapm_client.events[constants.TRANSACTION][2]["context"]["response"]
    assert elasticapm_client.events[constants.TRANSACTION][2]["context"]["request"]["body"] == "[REDACTED]"
    assert "headers" not in elasticapm_client.events[constants.ERROR][1]["context"]["request"]
    assert elasticapm_client.events[constants.ERROR][1]["context"]["request"]["body"] == "[REDACTED]"


def test_starlette_transaction_ignore_urls(app, elasticapm_client):
    client = TestClient(app)
    response = client.get("/hello")
    assert len(elasticapm_client.events[constants.TRANSACTION]) == 1

    elasticapm_client.config.update(1, transaction_ignore_urls="/*ello,/world")
    response = client.get("/hello")
    assert len(elasticapm_client.events[constants.TRANSACTION]) == 1


def test_transaction_name_is_route(app, elasticapm_client):
    client = TestClient(app)

    response = client.get("/hi/shay")

    assert response.status_code == 200

    assert len(elasticapm_client.events[constants.TRANSACTION]) == 1
    transaction = elasticapm_client.events[constants.TRANSACTION][0]
    assert transaction["name"] == "GET /hi/{name}"
    assert transaction["context"]["request"]["url"]["pathname"] == "/hi/shay"


@pytest.mark.skipif(starlette_version_tuple < (0, 14), reason="trailing slash behaviour new in 0.14")
@pytest.mark.parametrize(
    "url,expected",
    (
        ("/hi/shay/with/slash", "GET /hi/{name}/with/slash"),
        ("/hi/shay/without/slash/", "GET /hi/{name}/without/slash/"),
        ("/sub/subsub/hihi/shay/", "GET /sub/subsub/hihi/{name}/"),
    ),
)
def test_trailing_slash_redirect_detection(app, elasticapm_client, url, expected):
    client = TestClient(app)
    kwargs = {"allow_redirects": False} if starlette_version_tuple < (0, 43) else {"follow_redirects": False}
    response = client.get(url, **kwargs)
    assert response.status_code == 307
    assert len(elasticapm_client.events[constants.TRANSACTION]) == 1
    for transaction in elasticapm_client.events[constants.TRANSACTION]:
        assert transaction["name"] == expected


@pytest.mark.parametrize(
    "elasticapm_client",
    [
        {"enabled": False},
    ],
    indirect=True,
)
def test_enabled_instrumentation(app, elasticapm_client):
    client = TestClient(app)

    assert not isinstance(urllib3.connectionpool.HTTPConnectionPool.urlopen, wrapt.BoundFunctionWrapper)


def test_transaction_name_is_route_for_mounts(app, elasticapm_client):
    """
    Tests if recursive URL matching works when apps are mounted in other apps
    """
    client = TestClient(app)
    response = client.get("/sub/hi")
    assert response.status_code == 200

    assert len(elasticapm_client.events[constants.TRANSACTION]) == 1
    transaction = elasticapm_client.events[constants.TRANSACTION][0]
    assert transaction["name"] == "GET /sub/hi"
    assert transaction["context"]["request"]["url"]["pathname"] == "/sub/hi"

    response = client.get("/sub/subsub/hihi/shay")
    assert response.status_code == 200

    assert len(elasticapm_client.events[constants.TRANSACTION]) == 2
    transaction = elasticapm_client.events[constants.TRANSACTION][1]
    assert transaction["name"] == "GET /sub/subsub/hihi/{name}"
    assert transaction["context"]["request"]["url"]["pathname"] == "/sub/subsub/hihi/shay"


def test_undefined_route(app, elasticapm_client):
    client = TestClient(app)

    response = client.get("/undefined")

    assert response.status_code == 404
    assert len(elasticapm_client.events[constants.TRANSACTION]) == 1
    transaction = elasticapm_client.events[constants.TRANSACTION][0]
    assert transaction["name"] == "GET /undefined"


def test_undefined_mounted_route(app, elasticapm_client):
    client = TestClient(app)

    response = client.get("/sub/subsub/undefined")

    assert response.status_code == 404
    assert len(elasticapm_client.events[constants.TRANSACTION]) == 1
    transaction = elasticapm_client.events[constants.TRANSACTION][0]
    assert transaction["name"] == "GET /sub/subsub/undefined"


@pytest.mark.parametrize(
    "elasticapm_client",
    [
        {"capture_body": "error"},
    ],
    indirect=True,
)
def test_capture_body_error(app, elasticapm_client):
    """
    Context: https://github.com/elastic/apm-agent-python/issues/1032

    Before the above issue was fixed, this test would hang
    """
    client = TestClient(app)
    with pytest.raises(ValueError):
        response = client.post("/raise-exception", data="[0, 1]")


@pytest.fixture
def app_static_files_only(elasticapm_client):
    app = Starlette()
    app.add_middleware(ElasticAPM, client=elasticapm_client)
    app.mount("/tmp", StaticFiles(directory=file_path), name="static")

    yield app

    elasticapm.uninstrument()


def test_static_files_only(app_static_files_only, elasticapm_client):
    client = TestClient(app_static_files_only)

    response = client.get(
        "/tmp/" + file_name,
        headers={
            constants.TRACEPARENT_HEADER_NAME: "00-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-03",
            constants.TRACESTATE_HEADER_NAME: "foo=bar,bar=baz",
            "REMOTE_ADDR": "127.0.0.1",
        },
    )

    assert response.status_code == 200

    assert len(elasticapm_client.events[constants.TRANSACTION]) == 1
    transaction = elasticapm_client.events[constants.TRANSACTION][0]
    spans = elasticapm_client.spans_for_transaction(transaction)
    assert len(spans) == 0

    assert transaction["name"] == "GET /tmp"
    assert transaction["result"] == "HTTP 2xx"
    assert transaction["outcome"] == "success"
    assert transaction["type"] == "request"
    assert transaction["span_count"]["started"] == 0
    assert transaction["context"]["request"]["url"]["pathname"] == "/tmp/" + file_name
    request = transaction["context"]["request"]
    assert request["method"] == "GET"
    assert request["socket"] == {"remote_address": "127.0.0.1"}


def test_non_utf_8_body_in_ignored_paths_with_capture_body(app, elasticapm_client):
    client = TestClient(app)
    elasticapm_client.config.update(1, capture_body="all", transaction_ignore_urls="/hello")
    response = client.post("/hello", data=b"b$\x19\xc2")
    assert response.status_code == 200
    assert len(elasticapm_client.events[constants.TRANSACTION]) == 0


@pytest.mark.parametrize("elasticapm_client", [{"capture_body": "all"}], indirect=True)
def test_long_body(app, elasticapm_client):
    client = TestClient(app)

    response = client.post(
        "/",
        data={"foo": "b" * 10000},
    )

    assert response.status_code == 200

    assert len(elasticapm_client.events[constants.TRANSACTION]) == 1
    transaction = elasticapm_client.events[constants.TRANSACTION][0]
    request = transaction["context"]["request"]
    assert request["body"] == "foo=" + "b" * 9993 + "..."
    assert response.text == "10004"


def test_static_files_only_file_notfound(app_static_files_only, elasticapm_client):
    client = TestClient(app_static_files_only)

    response = client.get(
        "/tmp/whatever",
        headers={
            constants.TRACEPARENT_HEADER_NAME: "00-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-03",
            constants.TRACESTATE_HEADER_NAME: "foo=bar,bar=baz",
            "REMOTE_ADDR": "127.0.0.1",
        },
    )

    assert response.status_code == 404

    assert len(elasticapm_client.events[constants.TRANSACTION]) == 1
    transaction = elasticapm_client.events[constants.TRANSACTION][0]
    spans = elasticapm_client.spans_for_transaction(transaction)
    assert len(spans) == 0

    assert transaction["name"] == "GET /tmp"
    assert transaction["result"] == "HTTP 4xx"
    assert transaction["outcome"] == "success"
    assert transaction["type"] == "request"
    assert transaction["span_count"]["started"] == 0
    assert transaction["context"]["request"]["url"]["pathname"] == "/tmp/whatever"
    request = transaction["context"]["request"]
    assert request["method"] == "GET"
    assert request["socket"] == {"remote_address": "127.0.0.1"}


def test_make_client_with_config():
    c = make_apm_client(config={"SERVICE_NAME": "foo"}, client_cls=TempStoreClient)
    c.close()
    assert c.config.service_name == "foo"


def test_make_client_without_config():
    with mock.patch.dict("os.environ", {"ELASTIC_APM_SERVICE_NAME": "foo"}):
        c = make_apm_client(client_cls=TempStoreClient)
        c.close()
        assert c.config.service_name == "foo"


def test_websocket(app, elasticapm_client):
    client = TestClient(app)
    with client.websocket_connect("/ws") as websocket:
        data = websocket.receive_text()
        assert data == "Hello, world!"

    assert len(elasticapm_client.events[constants.TRANSACTION]) == 0


def test_transaction_active_in_base_exception_handler(app, elasticapm_client):
    client = TestClient(app)
    try:
        response = client.get("/raise-base-exception")
    except Exception as exc:
        # This is set by the exception handler -- we want to make sure the
        # handler has access to the transaction.
        assert exc.transaction_id

    assert len(elasticapm_client.events[constants.TRANSACTION]) == 1


def test_middleware_without_client_arg():
    with mock.patch.dict("os.environ", {"ELASTIC_APM_SERVICE_NAME": "foo"}):
        app = Starlette()
        elasticapm = ElasticAPM(app)

    assert elasticapm.client.config.service_name == "foo"



#  BSD 3-Clause License
#
#  Copyright (c) 2021, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import pytest

import elasticapm
from elasticapm.conf import constants


@pytest.mark.parametrize("elasticapm_client", [{"transaction_max_spans": 5}], indirect=True)
def test_transaction_max_spans(elasticapm_client):
    elasticapm_client.begin_transaction("test_type")
    for i in range(5):
        with elasticapm.capture_span("nodrop"):
            pass
    for i in range(10):
        with elasticapm.capture_span("drop"):
            pass
    transaction_obj = elasticapm_client.end_transaction("test")

    transaction = elasticapm_client.events[constants.TRANSACTION][0]
    spans = elasticapm_client.events[constants.SPAN]
    assert all(span["transaction_id"] == transaction["id"] for span in spans)

    assert transaction_obj.dropped_spans == 10
    assert len(spans) == 5
    for span in spans:
        assert span["name"] == "nodrop"
    assert transaction["span_count"] == {"dropped": 10, "started": 5}


@pytest.mark.parametrize("elasticapm_client", [{"transaction_max_spans": 3}], indirect=True)
def test_transaction_max_span_nested(elasticapm_client):
    elasticapm_client.begin_transaction("test_type")
    with elasticapm.capture_span("1"):
        with elasticapm.capture_span("2"):
            with elasticapm.capture_span("3"):
                with elasticapm.capture_span("4"):
                    with elasticapm.capture_span("5"):
                        pass
                with elasticapm.capture_span("6"):
                    pass
            with elasticapm.capture_span("7"):
                pass
        with elasticapm.capture_span("8"):
            pass
    with elasticapm.capture_span("9"):
        pass
    transaction_obj = elasticapm_client.end_transaction("test")

    transaction = elasticapm_client.events[constants.TRANSACTION][0]
    spans = elasticapm_client.events[constants.SPAN]

    assert transaction_obj.dropped_spans == 6
    assert len(spans) == 3
    for span in spans:
        assert span["name"] in ("1", "2", "3")
    assert transaction["span_count"] == {"dropped": 6, "started": 3}


@pytest.mark.parametrize("elasticapm_client", [{"transaction_max_spans": 1}], indirect=True)
def test_transaction_max_span_dropped_statistics(elasticapm_client):
    elasticapm_client.begin_transaction("test_type")
    with elasticapm.capture_span("not_dropped"):
        pass
    for i in range(10):
        resource = str(i % 2)
        with elasticapm.capture_span(
            span_type="x",
            span_subtype="y",
            extra={
                "destination": {"service": {"resource": resource}},
                "service": {"target": {"name": "foo", "type": "bar"}},
            },
            duration=100,
        ):
            pass
    elasticapm_client.end_transaction()
    transaction = elasticapm_client.events[constants.TRANSACTION][0]
    spans = elasticapm_client.events[constants.SPAN]

    assert len(spans) == 1
    for entry in transaction["dropped_spans_stats"]:
        assert entry["duration"]["count"] == 5
        assert entry["duration"]["sum"]["us"] == 500000000
        assert entry["service_target_type"] == "bar"
        assert entry["service_target_name"] == "foo"


@pytest.mark.parametrize(
    "elasticapm_client", [{"transaction_max_spans": 1, "server_version": (7, 15, 0)}], indirect=True
)
def test_transaction_max_span_dropped_statistics_not_collected_for_incompatible_server(elasticapm_client):
    elasticapm_client.begin_transaction("test_type")
    with elasticapm.capture_span("not_dropped"):
        pass
    with elasticapm.capture_span(
        span_type="x", span_subtype="x", extra={"destination": {"service": {"resource": "y"}}}, duration=100
    ):
        pass
    elasticapm_client.end_transaction()
    transaction = elasticapm_client.events[constants.TRANSACTION][0]
    spans = elasticapm_client.events[constants.SPAN]
    assert len(spans) == 1
    assert "dropped_spans_stats" not in transaction


@pytest.mark.parametrize("elasticapm_client", [{"exit_span_min_duration": "1ms"}], indirect=True)
def test_transaction_fast_exit_span(elasticapm_client):
    elasticapm_client.begin_transaction("test_type")
    with elasticapm.capture_span(span_type="x", name="x", leaf=True, duration=0.002):  # not dropped, too long
        pass
    with elasticapm.capture_span(span_type="y", name="y", leaf=True, duration=0.0001):  # dropped
        pass
    with elasticapm.capture_span(span_type="z", name="z", leaf=False, duration=0.0001):  # not dropped, not exit
        pass
    elasticapm_client.end_transaction("foo", duration=2.2)
    transaction = elasticapm_client.events[constants.TRANSACTION][0]
    spans = elasticapm_client.events[constants.SPAN]
    breakdown = elasticapm_client.metrics.get_metricset("elasticapm.metrics.sets.breakdown.BreakdownMetricSet")
    metrics = list(breakdown.collect())
    assert len(spans) == 2
    assert transaction["span_count"]["started"] == 3
    assert transaction["span_count"]["dropped"] == 1
    assert metrics[0]["span"]["type"] == "x"
    assert metrics[0]["samples"]["span.self_time.sum.us"]["value"] == 2000
    assert metrics[1]["span"]["type"] == "y"
    assert metrics[1]["samples"]["span.self_time.sum.us"]["value"] == 100


def test_transaction_cancelled_span(elasticapm_client):
    elasticapm_client.begin_transaction("test_type")
    with elasticapm.capture_span("test") as span:
        span.cancel()
    elasticapm_client.end_transaction("foo")
    transaction = elasticapm_client.events[constants.TRANSACTION][0]
    spans = elasticapm_client.events[constants.SPAN]
    assert len(spans) == 0
    assert transaction["span_count"]["started"] == 0
    assert transaction["span_count"]["dropped"] == 0



# -*- coding: utf-8 -*-

#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import hashlib
import json
import os
import re
import ssl
import urllib.parse
from urllib.request import getproxies_environment, proxy_bypass_environment

import urllib3
from urllib3.exceptions import MaxRetryError, TimeoutError

from elasticapm.transport.exceptions import TransportException
from elasticapm.transport.http_base import HTTPTransportBase
from elasticapm.utils import json_encoder, read_pem_file
from elasticapm.utils.logging import get_logger

try:
    import certifi
except ImportError:
    certifi = None

logger = get_logger("elasticapm.transport.http")


class Transport(HTTPTransportBase):
    def __init__(self, url: str, *args, **kwargs) -> None:
        super(Transport, self).__init__(url, *args, **kwargs)
        pool_kwargs = {"cert_reqs": "CERT_REQUIRED", "ca_certs": self.ca_certs, "block": True}
        if url.startswith("https"):
            if self._server_cert:
                pool_kwargs.update(
                    {"assert_fingerprint": self.cert_fingerprint, "assert_hostname": False, "cert_reqs": ssl.CERT_NONE}
                )
                del pool_kwargs["ca_certs"]
            elif not self._verify_server_cert:
                pool_kwargs["cert_reqs"] = ssl.CERT_NONE
                pool_kwargs["assert_hostname"] = False
        self._pool_kwargs = pool_kwargs
        self._http = None
        self._url = url

    def send(self, data, forced_flush=False, custom_url=None, custom_headers=None):
        response = None

        headers = self._headers.copy() if self._headers else {}
        headers.update(self.auth_headers)
        if custom_headers:
            headers.update(custom_headers)
        else:
            headers.update(
                {
                    b"Content-Type": b"application/x-ndjson",
                    b"Content-Encoding": b"gzip",
                }
            )

        url = custom_url or self._url
        if forced_flush:
            url = f"{url}?flushed=true"
        try:
            try:
                response = self.http.urlopen(
                    "POST", url, body=data, headers=headers, timeout=self._timeout, preload_content=False
                )
                logger.debug("Sent request, url=%s size=%.2fkb status=%s", url, len(data) / 1024.0, response.status)
            except Exception as e:
                if isinstance(e, MaxRetryError) and isinstance(e.reason, TimeoutError):
                    message = "Connection to APM Server timed out " "(url: %s, timeout: %s seconds)" % (
                        self._url,
                        self._timeout,
                    )
                else:
                    message = "Unable to reach APM Server: %s (url: %s)" % (e, self._url)
                raise TransportException(message, data)
            body = response.read()
            if response.status >= 400:
                if response.status == 429:  # rate-limited
                    message = "Temporarily rate limited: "
                else:
                    message = "HTTP %s: " % response.status
                message += body.decode("utf8", errors="replace")[:10000]
                raise TransportException(message, data)
            return response.headers.get("Location")
        finally:
            if response:
                response.close()

    @property
    def http(self) -> urllib3.PoolManager:
        if not self._http:
            url_parts = urllib.parse.urlparse(self._url)
            proxies = getproxies_environment()
            proxy_url = proxies.get("https", proxies.get("http", None))
            if proxy_url and not proxy_bypass_environment(url_parts.netloc):
                self._http = urllib3.ProxyManager(proxy_url, **self._pool_kwargs)
            else:
                self._http = urllib3.PoolManager(**self._pool_kwargs)
        return self._http

    def handle_fork(self) -> None:
        # reset http pool to avoid sharing connections with the parent process
        self._http = None

    def get_config(self, current_version=None, keys=None):
        """
        Gets configuration from a remote APM Server

        :param current_version: version of the current configuration
        :param keys: a JSON-serializable dict to identify this instance, e.g.
                {
                    "service": {
                        "name": "foo",
                        "environment": "bar"
                    }
                }
        :return: a three-tuple of new version, config dictionary and validity in seconds.
            Any element of the tuple can be None.
        """
        url = self._config_url
        data = json_encoder.dumps(keys).encode("utf-8")
        headers = self._headers.copy()
        headers[b"Content-Type"] = "application/json"
        headers.update(self.auth_headers)
        max_age = 300
        if current_version:
            headers["If-None-Match"] = current_version
        try:
            response = self.http.urlopen(
                "POST", url, body=data, headers=headers, timeout=self._timeout, preload_content=False
            )
        except (urllib3.exceptions.RequestError, urllib3.exceptions.HTTPError) as e:
            logger.debug("HTTP error while fetching remote config: %s", str(e))
            return current_version, None, max_age
        body = response.read()

        max_age = self._get_cache_control_max_age(response.headers) or max_age

        if response.status == 304:
            # config is unchanged, return
            logger.debug("Configuration unchanged")
            return current_version, None, max_age
        elif response.status >= 400:
            return None, None, max_age

        if not body:
            logger.debug("APM Server answered with empty body and status code %s", response.status)
            return current_version, None, max_age
        body = body.decode("utf-8")
        try:
            data = json_encoder.loads(body)
            return response.headers.get("Etag"), data, max_age
        except json.JSONDecodeError:
            logger.warning("Failed decoding APM Server response as JSON: %s", body)
            return current_version, None, max_age

    def _get_cache_control_max_age(self, response_headers):
        max_age = None
        if "Cache-Control" in response_headers:
            try:
                cc_max_age = int(next(re.finditer(r"max-age=(\d+)", response_headers["Cache-Control"])).groups()[0])
                if cc_max_age <= 0:
                    # max_age remains at default value
                    pass
                elif cc_max_age < 5:
                    max_age = 5
                else:
                    max_age = cc_max_age
            except StopIteration:
                logger.debug("Could not parse Cache-Control header: %s", response_headers["Cache-Control"])
        return max_age

    def _process_queue(self) -> None:
        if not self.client.server_version:
            # this is useful on aws lambda environments where this call incurs in unwanted latency
            if self.client.config.skip_server_info:
                logger.debug("Skipping to fetch server info")
            else:
                self.fetch_server_info()
        super()._process_queue()

    def fetch_server_info(self) -> None:
        headers = self._headers.copy() if self._headers else {}
        headers.update(self.auth_headers)
        headers[b"accept"] = b"text/plain"
        try:
            response = self.http.urlopen("GET", self._server_info_url, headers=headers, timeout=self._timeout)
            body = response.data
            data = json_encoder.loads(body.decode("utf8"))
            version = data["version"]
            logger.debug("Fetched APM Server version %s", version)
            self.client.server_version = version_string_to_tuple(version)
        except (urllib3.exceptions.RequestError, urllib3.exceptions.HTTPError) as e:
            logger.debug("HTTP error while fetching server information: %s", str(e))
        except json.JSONDecodeError as e:
            logger.debug(
                f"JSON decoding error while fetching server information. Error: {str(e)} Body: {body.decode('utf8')}"
            )
        except (KeyError, TypeError):
            logger.debug("No version key found in server response: %s", response.data)

    @property
    def cert_fingerprint(self):
        if self._server_cert:
            with open(self._server_cert, "rb") as f:
                cert_data = read_pem_file(f)
            digest = hashlib.sha256()
            digest.update(cert_data)
            return digest.hexdigest()
        return None

    @property
    def auth_headers(self):
        headers = super(Transport, self).auth_headers
        return {k.encode("ascii"): v.encode("ascii") for k, v in headers.items()}

    @property
    def ca_certs(self):
        """
        Return location of certificate store. If the server_ca_cert_file config option is set,
        its value is returned. Otherwise, the certifi store is used, unless it is disabled or not installed.
        """
        if self._server_ca_cert_file:
            return self._server_ca_cert_file
        return certifi.where() if (certifi and self.client.config.use_certifi) else None

    def close(self):
        """
        Take care of being able to shutdown cleanly
        :return:
        """
        if self._closed or (not self._thread or self._thread.pid != os.getpid()):
            return

        self._closed = True
        # we are racing against urllib3 ConnectionPool weakref finalizer that would lead to having them closed
        # and we hanging waiting for send any eventual queued data
        # Force the creation of a new PoolManager so that we are always able to flush
        self._http = None
        self.queue("close", None)
        if not self._flushed.wait(timeout=self._max_flush_time_seconds):
            logger.error("Closing the transport connection timed out.")


def version_string_to_tuple(version):
    if version:
        version_parts = re.split(r"[.\-]", version)
        return tuple(int(p) if p.isdigit() else p for p in version_parts)
    return ()


# left for backwards compatibility
AsyncTransport = Transport



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from elasticapm.instrumentation.packages.dbapi2 import (
    ConnectionProxy,
    CursorProxy,
    DbApi2Instrumentation,
    extract_signature,
)
from elasticapm.utils import default_ports


class PyMSSQLCursorProxy(CursorProxy):
    provider_name = "mssql"

    def extract_signature(self, sql):
        return extract_signature(sql)


class PyMSSQLConnectionProxy(ConnectionProxy):
    cursor_proxy = PyMSSQLCursorProxy

    def cursor(self, *args, **kwargs):
        c = super().cursor(*args, **kwargs)
        c._self_database = self._self_database
        return c


class PyMSSQLInstrumentation(DbApi2Instrumentation):
    name = "pymssql"

    instrument_list = [("pymssql", "connect")]

    def call(self, module, method, wrapped, instance, args, kwargs):
        host, port = get_host_port(args, kwargs)
        destination_info = {
            "address": host,
            "port": port,
        }
        db = args[3] if len(args) > 3 else kwargs.get("database", "")
        proxy = PyMSSQLConnectionProxy(wrapped(*args, **kwargs), destination_info=destination_info)
        proxy._self_database = db
        return proxy


def get_host_port(args, kwargs):
    host = args[0] if args else kwargs.get("server")
    port = None
    if not host:
        host = kwargs.get("host", "localhost")
        for sep in (",", ":"):
            if sep in host:
                host, port = host.rsplit(sep, 1)
                port = int(port)
                break
    if not port:
        port = int(kwargs.get("port", default_ports.get("mssql")))
    return host, port



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE

import platform
import time

import mock
import pytest

from elasticapm.utils.threading import IntervalTimer
from tests.utils import assert_any_record_contains


def test_interval_timer():
    func = mock.Mock()
    timer = IntervalTimer(function=func, interval=0.1, args=(1,), kwargs={"a": "b"})
    timer.start()
    time.sleep(0.25)
    try:
        assert func.call_count == 2
        for call in func.call_args_list:
            assert call == ((1,), {"a": "b"})
    finally:
        timer.cancel()
    time.sleep(0.05)
    assert not timer.is_alive()


@pytest.mark.skipif(platform.system() == "Windows", reason="Flaky test on windows")
def test_interval_timer_interval_override():
    func = mock.Mock()
    func.return_value = 0.05
    timer = IntervalTimer(function=func, interval=0.1, evaluate_function_interval=True)
    timer.start()
    time.sleep(0.25)
    try:
        assert func.call_count in (3, 4)
    finally:
        timer.cancel()
    time.sleep(0.05)
    assert not timer.is_alive()


def test_interval_timer_interval_override_non_number():
    func = mock.Mock()
    func.return_value = "foo"
    timer = IntervalTimer(function=func, interval=0.1, evaluate_function_interval=True)
    timer.start()
    time.sleep(0.25)
    try:
        assert func.call_count == 2
    finally:
        timer.cancel()
    time.sleep(0.05)
    assert not timer.is_alive()


def test_interval_timer_exception(caplog):
    def my_func():
        return 1 / 0

    with caplog.at_level("ERROR", "elasticapm.utils.threading"):
        timer = IntervalTimer(function=my_func, interval=0.1)
        timer.start()
        time.sleep(0.25)
        assert timer.is_alive()
    timer.cancel()
    assert_any_record_contains(caplog.records, "Exception in interval timer function")



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os

import mock
import pytest

import elasticapm
from elasticapm.conf.constants import ERROR, KEYWORD_MAX_LENGTH
from elasticapm.utils import encoding
from tests.utils.stacks import get_me_more_test_frames


def test_explicit_message_on_exception_event(elasticapm_client):
    try:
        raise ValueError("foo")
    except ValueError:
        elasticapm_client.capture("Exception", message="foobar")

    assert len(elasticapm_client.events) == 1
    event = elasticapm_client.events[ERROR][0]
    assert event["exception"]["message"] == "foobar"


@pytest.mark.parametrize(
    "elasticapm_client",
    [{"include_paths": ("tests",), "local_var_max_length": 20, "local_var_list_max_length": 10}],
    indirect=True,
)
def test_exception_event(elasticapm_client):
    try:
        a_local_var = 1
        a_long_local_var = 100 * "a"
        a_long_local_list = list(range(100))
        raise ValueError("foo")
    except ValueError:
        elasticapm_client.capture("Exception")

    assert len(elasticapm_client.events) == 1
    event = elasticapm_client.events[ERROR][0]
    assert "exception" in event
    exc = event["exception"]
    assert exc["message"] == "ValueError: foo"
    assert exc["type"] == "ValueError"
    assert exc["module"] == ValueError.__module__  # this differs in some Python versions
    assert "stacktrace" in exc
    frames = exc["stacktrace"]
    assert len(frames) == 1
    frame = frames[0]
    assert frame["abs_path"], __file__.replace(".pyc" == ".py")
    assert frame["filename"] == os.path.join("tests", "client", "exception_tests.py")
    assert frame["module"] == __name__
    assert frame["function"] == "test_exception_event"
    assert not frame["library_frame"]
    assert frame["vars"]["a_local_var"] == 1
    assert len(frame["vars"]["a_long_local_var"]) == 20
    assert len(frame["vars"]["a_long_local_list"]) == 12
    assert frame["vars"]["a_long_local_list"][-1] == "(90 more elements)"
    assert "timestamp" in event
    assert "log" not in event
    # check that only frames from `tests` module are not marked as library frames
    assert all(
        frame["library_frame"] or frame["module"].startswith("tests") for frame in event["exception"]["stacktrace"]
    )


def test_sending_exception(sending_elasticapm_client: elasticapm.Client):
    try:
        1 / 0
    except Exception:
        sending_elasticapm_client.capture_exception()
    sending_elasticapm_client.close()
    assert (
        sending_elasticapm_client.httpserver.responses[0]["code"] == 202
    ), sending_elasticapm_client.httpserver.responses[0]


@pytest.mark.parametrize(
    "elasticapm_client",
    [{"include_paths": ("*/tests/*",), "local_var_max_length": 20, "local_var_list_max_length": 10}],
    indirect=True,
)
def test_message_event(elasticapm_client):
    a_local_var = 1
    a_long_local_var = 100 * "a"
    a_long_local_list = list(range(100))
    elasticapm_client.capture("Message", message="test")

    assert len(elasticapm_client.events) == 1
    event = elasticapm_client.events[ERROR][0]
    assert event["log"]["message"] == "test"
    assert "stacktrace" not in event
    assert "timestamp" in event
    assert "stacktrace" in event["log"]
    # check that only frames from `tests` module are not marked as library frames
    for frame in event["log"]["stacktrace"]:
        # vscode's python extension runs tests using libraries from the
        # extension's dir, which results in false positives here. This is where
        # runpy, _pydevd, and debugpy come from.
        assert frame["library_frame"] or frame["module"].startswith(
            ("tests", "__main__", "runpy", "_pydevd", "debugpy")
        ), (
            frame["module"],
            frame["abs_path"],
        )
    frame = event["log"]["stacktrace"][0]
    assert frame["vars"]["a_local_var"] == 1
    assert len(frame["vars"]["a_long_local_var"]) == 20
    assert len(frame["vars"]["a_long_local_list"]) == 12
    assert frame["vars"]["a_long_local_list"][-1] == "(90 more elements)"


def test_param_message_event(elasticapm_client):
    elasticapm_client.capture("Message", param_message={"message": "test %s %d", "params": ("x", 1)})

    assert len(elasticapm_client.events[ERROR]) == 1
    event = elasticapm_client.events[ERROR][0]
    assert event["log"]["message"] == "test x 1"
    assert event["log"]["param_message"] == "test %s %d"


def test_message_with_percent(elasticapm_client):
    elasticapm_client.capture("Message", message="This works 100% of the time")

    assert len(elasticapm_client.events[ERROR]) == 1
    event = elasticapm_client.events[ERROR][0]
    assert event["log"]["message"] == "This works 100% of the time"
    assert event["log"]["param_message"] == "This works 100% of the time"


def test_logger(elasticapm_client):
    elasticapm_client.capture("Message", message="test", logger_name="test")

    assert len(elasticapm_client.events[ERROR]) == 1
    event = elasticapm_client.events[ERROR][0]
    assert event["log"]["logger_name"] == "test"
    assert "timestamp" in event


@pytest.mark.parametrize(
    "elasticapm_client",
    [
        {"collect_local_variables": "errors"},
        {"collect_local_variables": "transactions"},
        {"collect_local_variables": "all"},
        {"collect_local_variables": "something"},
    ],
    indirect=True,
)
def test_collect_local_variables_errors(elasticapm_client):
    mode = elasticapm_client.config.collect_local_variables
    try:
        1 / 0
    except ZeroDivisionError:
        elasticapm_client.capture_exception()
    event = elasticapm_client.events[ERROR][0]
    if mode in ("errors", "all"):
        assert "vars" in event["exception"]["stacktrace"][0], mode
    else:
        assert "vars" not in event["exception"]["stacktrace"][0], mode


@pytest.mark.parametrize(
    "elasticapm_client",
    [
        {"source_lines_error_library_frames": 0, "source_lines_error_app_frames": 0},
        {"source_lines_error_library_frames": 1, "source_lines_error_app_frames": 1},
        {"source_lines_error_library_frames": 7, "source_lines_error_app_frames": 3},
    ],
    indirect=True,
)
def test_collect_source_errors(elasticapm_client):
    library_frame_context = elasticapm_client.config.source_lines_error_library_frames
    in_app_frame_context = elasticapm_client.config.source_lines_error_app_frames
    try:
        import datetime
        import json

        json.dumps(datetime.datetime.now())
    except TypeError:
        elasticapm_client.capture_exception()
    event = elasticapm_client.events[ERROR][0]
    in_app_frame = event["exception"]["stacktrace"][0]
    library_frame = event["exception"]["stacktrace"][1]
    assert not in_app_frame["library_frame"]
    assert library_frame["library_frame"]
    if library_frame_context:
        assert "context_line" in library_frame, library_frame_context
        assert "pre_context" in library_frame, library_frame_context
        assert "post_context" in library_frame, library_frame_context
        lines = len([library_frame["context_line"]] + library_frame["pre_context"] + library_frame["post_context"])
        assert lines == library_frame_context, library_frame_context
    else:
        assert "context_line" not in library_frame, library_frame_context
        assert "pre_context" not in library_frame, library_frame_context
        assert "post_context" not in library_frame, library_frame_context
    if in_app_frame_context:
        assert "context_line" in in_app_frame, in_app_frame_context
        assert "pre_context" in in_app_frame, in_app_frame_context
        assert "post_context" in in_app_frame, in_app_frame_context
        lines = len([in_app_frame["context_line"]] + in_app_frame["pre_context"] + in_app_frame["post_context"])
        assert lines == in_app_frame_context, (in_app_frame_context, in_app_frame["lineno"])
    else:
        assert "context_line" not in in_app_frame, in_app_frame_context
        assert "pre_context" not in in_app_frame, in_app_frame_context
        assert "post_context" not in in_app_frame, in_app_frame_context


def test_transaction_data_is_attached_to_errors_no_transaction(elasticapm_client):
    elasticapm_client.capture_message("noid")
    elasticapm_client.begin_transaction("test")
    elasticapm_client.end_transaction("test", "test")
    elasticapm_client.capture_message("noid")

    errors = elasticapm_client.events[ERROR]
    assert "transaction_id" not in errors[0]
    assert "transaction_id" not in errors[1]


def test_transaction_data_is_attached_to_errors_message_outside_span(elasticapm_client):
    elasticapm_client.begin_transaction("test")
    elasticapm_client.capture_message("outside_span")
    transaction = elasticapm_client.end_transaction("test", "test")

    error = elasticapm_client.events[ERROR][0]
    assert error["transaction_id"] == transaction.id
    assert error["parent_id"] == transaction.id
    assert error["transaction"]["sampled"]
    assert error["transaction"]["type"] == "test"


def test_transaction_data_is_attached_to_errors_message_in_span(elasticapm_client):
    elasticapm_client.begin_transaction("test")

    with elasticapm.capture_span("in_span_handler_test") as span_obj:
        elasticapm_client.capture_message("in_span")

    transaction = elasticapm_client.end_transaction("test", "test")

    error = elasticapm_client.events[ERROR][0]

    assert error["transaction_id"] == transaction.id
    assert error["parent_id"] == span_obj.id
    assert error["transaction"]["sampled"]
    assert error["transaction"]["type"] == "test"


def test_transaction_data_is_attached_to_errors_exc_handled_in_span(elasticapm_client):
    elasticapm_client.begin_transaction("test")
    with elasticapm.capture_span("in_span_handler_test") as span_obj:
        try:
            assert False
        except AssertionError:
            elasticapm_client.capture_exception()
    transaction = elasticapm_client.end_transaction("test", "test")

    error = elasticapm_client.events[ERROR][0]

    assert error["transaction_id"] == transaction.id
    assert error["parent_id"] == span_obj.id
    assert error["transaction"]["sampled"]
    assert error["transaction"]["type"] == "test"


def test_transaction_data_is_attached_to_errors_exc_handled_outside_span(elasticapm_client):
    elasticapm_client.begin_transaction("test")
    try:
        with elasticapm.capture_span("out_of_span_handler_test") as span_obj:
            assert False
    except AssertionError:
        elasticapm_client.capture_exception()
    transaction = elasticapm_client.end_transaction("test", "test")

    error = elasticapm_client.events[ERROR][0]

    assert error["transaction_id"] == transaction.id
    assert error["parent_id"] == span_obj.id
    assert error["transaction"]["sampled"]
    assert error["transaction"]["type"] == "test"


def test_transaction_context_is_used_in_errors(elasticapm_client):
    elasticapm_client.begin_transaction("test")
    elasticapm.label(foo="baz")
    elasticapm.set_custom_context({"a": "b"})
    elasticapm.set_user_context(username="foo", email="foo@example.com", user_id=42)
    elasticapm_client.capture_message("x", custom={"foo": "bar"})
    transaction = elasticapm_client.end_transaction("test", "OK")
    message = elasticapm_client.events[ERROR][0]
    assert message["context"]["custom"] == {"a": "b", "foo": "bar"}
    assert message["context"]["user"] == {"username": "foo", "email": "foo@example.com", "id": 42}
    assert message["context"]["tags"] == {"foo": "baz"}
    assert "a" in transaction.context["custom"]
    assert "foo" not in transaction.context["custom"]


def test_error_keyword_truncation(sending_elasticapm_client):
    too_long = "x" * (KEYWORD_MAX_LENGTH + 1)
    expected = encoding.keyword_field(too_long)

    # let's create a way too long Exception type with a way too long module name
    WayTooLongException = type(too_long.upper(), (Exception,), {})
    WayTooLongException.__module__ = too_long
    try:
        raise WayTooLongException()
    except WayTooLongException:
        with mock.patch("elasticapm.events.get_culprit") as mock_get_culprit:
            mock_get_culprit.return_value = too_long
            sending_elasticapm_client.capture_exception(handled=False)
    sending_elasticapm_client.close()
    error = sending_elasticapm_client.httpserver.payloads[0][1]["error"]

    assert error["exception"]["type"] == expected.upper()
    assert error["exception"]["module"] == expected
    assert error["culprit"] == expected


def test_message_keyword_truncation(sending_elasticapm_client):
    too_long = "x" * (KEYWORD_MAX_LENGTH + 1)
    expected = encoding.keyword_field(too_long)
    sending_elasticapm_client.capture_message(
        param_message={"message": too_long, "params": []}, logger_name=too_long, handled=False
    )
    sending_elasticapm_client.close()
    error = sending_elasticapm_client.httpserver.payloads[0][1]["error"]

    assert error["log"]["param_message"] == expected
    assert error["log"]["message"] == too_long  # message is not truncated

    assert error["log"]["logger_name"] == expected


@pytest.mark.parametrize("elasticapm_client", [{"stack_trace_limit": 10}], indirect=True)
def test_stack_trace_limit(elasticapm_client):
    def func():
        1 / 0  # I'm the context line of the last frame!

    try:
        list(get_me_more_test_frames(15, func))
    except ZeroDivisionError:
        elasticapm_client.capture_exception()
    exception = elasticapm_client.events[ERROR][-1]
    frames = exception["exception"]["stacktrace"]
    assert len(frames) == 10
    assert "I'm the context line of the last frame" in frames[-1]["context_line"]

    elasticapm_client.config.update("1", stack_trace_limit=-1)
    try:
        list(get_me_more_test_frames(15, func))
    except ZeroDivisionError:
        elasticapm_client.capture_exception()
    exception = elasticapm_client.events[ERROR][-1]
    frames = exception["exception"]["stacktrace"]
    assert len(frames) > 15
    assert "I'm the context line of the last frame" in frames[-1]["context_line"]

    elasticapm_client.config.update("1", stack_trace_limit=0)
    try:
        list(get_me_more_test_frames(15, func))
    except ZeroDivisionError:
        elasticapm_client.capture_exception()
    exception = elasticapm_client.events[ERROR][-1]
    frames = exception["exception"]["stacktrace"]
    assert len(frames) == 0


def test_fail_on_uuid_raise(elasticapm_client):
    def generate_uuid():
        from uuid import UUID

        return UUID("INVALID")

    try:
        generate_uuid()
    except Exception:
        elasticapm_client.capture_exception()


def test_long_message(elasticapm_client):
    try:
        raise Exception("t" * 1000000)
    except:
        elasticapm_client.capture_exception()

    exception = elasticapm_client.events[ERROR][0]
    assert exception["exception"]["message"] == f'Exception: {"t"*9988}{"…"}'



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import pytest  # isort:skip

sanic = pytest.importorskip("sanic")  # isort:skip

import json

from elasticapm.conf import constants

pytestmark = [pytest.mark.sanic]  # isort:skip


@pytest.mark.parametrize(
    "url, transaction_name, span_count, custom_context",
    [("/", "GET /", 1, {}), ("/greet/sanic", "GET /greet/<name:str>", 0, {"name": "sanic"})],
)
def test_get(url, transaction_name, span_count, custom_context, sanic_elastic_app, elasticapm_client):
    sanic_app, apm = next(sanic_elastic_app(elastic_client=elasticapm_client))
    if int(sanic.__version__.split(".")[0]) < 21 and url != "/":
        pytest.skip("str type doesn't work in Sanic 20.x.x")
    source_request, response = sanic_app.test_client.get(
        url,
        headers={
            constants.TRACEPARENT_HEADER_NAME: "00-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-03",
            constants.TRACESTATE_HEADER_NAME: "foo=bar,bar=baz",
            "REMOTE_ADDR": "127.0.0.1",
        },
    )  # type: sanic.response.HttpResponse

    assert len(elasticapm_client.events[constants.TRANSACTION]) == 1
    transaction = elasticapm_client.events[constants.TRANSACTION][0]
    spans = elasticapm_client.spans_for_transaction(transaction)
    assert len(spans) == span_count
    if span_count > 0:
        span = spans[0]
        assert span["name"] == "test"
        assert transaction["span_count"]["started"] == span_count

    for field, value in {
        "name": transaction_name,
        "result": "HTTP 2xx",
        "outcome": "success",
        "type": "request",
    }.items():
        assert transaction[field] == value

    request = transaction["context"]["request"]
    assert request["method"] == "GET"
    assert request["socket"] == {"remote_address": f"127.0.0.1", "encrypted": False}
    context = transaction["context"]["custom"]
    assert context == custom_context


def test_capture_exception(sanic_elastic_app, elasticapm_client):
    sanic_app, apm = next(sanic_elastic_app(elastic_client=elasticapm_client))
    _, _ = sanic_app.test_client.get(
        "/capture-exception",
        headers={
            constants.TRACEPARENT_HEADER_NAME: "00-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-03",
            constants.TRACESTATE_HEADER_NAME: "foo=bar,bar=baz",
            "REMOTE_ADDR": "127.0.0.1",
        },
    )

    assert len(elasticapm_client.events[constants.ERROR]) == 1
    transaction = elasticapm_client.events[constants.TRANSACTION][0]

    for field, value in {
        "name": "GET /capture-exception",
        "result": "HTTP 5xx",
        "outcome": "failure",
        "type": "request",
    }.items():
        assert transaction[field] == value


@pytest.mark.parametrize(
    "elasticapm_client", [{"capture_body": "errors"}, {"capture_body": "all"}, {"capture_body": "off"}], indirect=True
)
def test_capture_body(sanic_elastic_app, elasticapm_client):
    sanic_app, apm = next(sanic_elastic_app(elastic_client=elasticapm_client))
    _, resp = sanic_app.test_client.post(
        "/v1/apm/unhandled-exception",
        headers={
            constants.TRACEPARENT_HEADER_NAME: "00-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-03",
            constants.TRACESTATE_HEADER_NAME: "foo=bar,bar=baz",
            "REMOTE_ADDR": "127.0.0.1",
        },
        data=json.dumps({"foo": "bar"}),
    )
    assert len(elasticapm_client.events[constants.ERROR]) == 1
    transaction = elasticapm_client.events[constants.TRANSACTION][0]
    for field, value in {
        "name": "POST /v1/apm/unhandled-exception",
        "result": "HTTP 5xx",
        "outcome": "failure",
        "type": "request",
    }.items():
        assert transaction[field] == value
    if elasticapm_client.config.capture_body in ("all", "errors"):
        assert transaction["context"]["request"]["body"] == '{"foo": "bar"}'
    else:
        assert transaction["context"]["request"]["body"] == "[REDACTED]"


def test_unhandled_exception_capture(sanic_elastic_app, elasticapm_client):
    sanic_app, apm = next(sanic_elastic_app(elastic_client=elasticapm_client))
    _, resp = sanic_app.test_client.post(
        "/v1/apm/unhandled-exception",
        headers={
            constants.TRACEPARENT_HEADER_NAME: "00-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-03",
            constants.TRACESTATE_HEADER_NAME: "foo=bar,bar=baz",
            "REMOTE_ADDR": "127.0.0.1",
        },
    )
    assert len(elasticapm_client.events[constants.ERROR]) == 1
    transaction = elasticapm_client.events[constants.TRANSACTION][0]
    for field, value in {
        "name": "POST /v1/apm/unhandled-exception",
        "result": "HTTP 5xx",
        "outcome": "failure",
        "type": "request",
    }.items():
        assert transaction[field] == value


@pytest.mark.parametrize(
    "url, expected_source",
    [
        ("/raise-exception", "custom-handler"),
        ("/raise-value-error", "value-error-custom"),
        ("/fallback-value-error", "custom-handler-default"),
    ],
)
def test_client_with_custom_error_handler(
    url, expected_source, sanic_elastic_app, elasticapm_client, custom_error_handler
):
    sanic_app, apm = next(sanic_elastic_app(elastic_client=elasticapm_client, error_handler=custom_error_handler))
    _, resp = sanic_app.test_client.get(
        url,
        headers={
            constants.TRACEPARENT_HEADER_NAME: "00-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-03",
            constants.TRACESTATE_HEADER_NAME: "foo=bar,bar=baz",
            "REMOTE_ADDR": "127.0.0.1",
        },
    )
    assert len(elasticapm_client.events[constants.ERROR]) == 1
    assert resp.json["source"] == expected_source


def test_header_field_sanitization(sanic_elastic_app, elasticapm_client):
    sanic_app, apm = next(sanic_elastic_app(elastic_client=elasticapm_client))
    _, resp = sanic_app.test_client.get(
        "/add-custom-headers",
        headers={
            constants.TRACEPARENT_HEADER_NAME: "00-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-03",
            constants.TRACESTATE_HEADER_NAME: "foo=bar,bar=baz",
            "REMOTE_ADDR": "127.0.0.1",
            "API_KEY": "some-random-api-key",
        },
    )
    assert len(apm._client.events[constants.TRANSACTION]) == 1
    transaction = apm._client.events[constants.TRANSACTION][0]
    assert transaction["context"]["response"]["headers"]["sessionid"] == "[REDACTED]"
    assert transaction["context"]["request"]["headers"]["api_key"] == "[REDACTED]"


def test_cookies_normalization(sanic_elastic_app, elasticapm_client):
    sanic_app, apm = next(sanic_elastic_app(elastic_client=elasticapm_client))
    _, resp = sanic_app.test_client.get(
        "/add-cookies",
    )
    assert resp.status_code == 200
    assert len(apm._client.events[constants.TRANSACTION]) == 1
    transaction = apm._client.events[constants.TRANSACTION][0]
    assert transaction["context"]["response"]["cookies"] == {"some": {"value": "cookie", "path": "/"}}


def test_custom_callback_handlers(sanic_elastic_app, elasticapm_client):
    def _custom_transaction_callback(request):
        return "my-custom-name"

    async def _user_info_callback(request):
        return "test", "test@test.com", 1234356

    async def _label_callback(request):
        return {
            "label1": "value1",
            "label2": 19,
        }

    sanic_app, apm = next(
        sanic_elastic_app(
            elastic_client=elasticapm_client,
            transaction_name_callback=_custom_transaction_callback,
            user_context_callback=_user_info_callback,
            label_info_callback=_label_callback,
        )
    )
    _, resp = sanic_app.test_client.get(
        "/add-custom-headers",
        headers={
            constants.TRACEPARENT_HEADER_NAME: "00-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-03",
            constants.TRACESTATE_HEADER_NAME: "foo=bar,bar=baz",
            "REMOTE_ADDR": "127.0.0.1",
            "API_KEY": "some-random-api-key",
        },
    )
    assert len(apm._client.events[constants.TRANSACTION]) == 1
    assert apm._client.events[constants.TRANSACTION][0]["name"] == "my-custom-name"
    assert apm._client.events[constants.TRANSACTION][0]["context"]["user"]["username"] == "test"
    assert apm._client.events[constants.TRANSACTION][0]["context"]["user"]["id"] == 1234356
    assert apm._client.events[constants.TRANSACTION][0]["context"]["tags"]["label2"] == 19



#  BSD 3-Clause License
#
#  Copyright (c) 2022, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import pytest

grpc = pytest.importorskip("grpc")

from unittest.mock import MagicMock, patch

pytestmark = pytest.mark.grpc

from elasticapm.instrumentation.packages.grpc import (
    GRPCAsyncServerInstrumentation,
    GRPCClientInstrumentation,
    GRPCServerInstrumentation,
)

# ---------------------------------------------------------------------------
# Helpers
# ---------------------------------------------------------------------------


def _make_client_instrumentation():
    return GRPCClientInstrumentation()


def _make_server_instrumentation():
    return GRPCServerInstrumentation()


def _make_async_server_instrumentation():
    return GRPCAsyncServerInstrumentation()


# ---------------------------------------------------------------------------
# GRPCClientInstrumentation
# ---------------------------------------------------------------------------


class TestGRPCClientInstrumentation:
    def test_insecure_channel_positional_arg_parses_host_port(self):
        """insecure_channel called with 'host:port' as positional arg."""
        instrumentation = _make_client_instrumentation()
        fake_channel = MagicMock()
        wrapped = MagicMock(return_value=fake_channel)
        intercepted_channel = MagicMock()

        with patch("grpc.intercept_channel", return_value=intercepted_channel) as mock_intercept, patch(
            "elasticapm.contrib.grpc.client_interceptor._ClientInterceptor"
        ) as MockInterceptor:
            result = instrumentation.call(
                module="grpc",
                method="insecure_channel",
                wrapped=wrapped,
                instance=None,
                args=("myhost:50051",),
                kwargs={},
            )

        wrapped.assert_called_once_with("myhost:50051")
        MockInterceptor.assert_called_once_with("myhost", 50051, secure=False)
        mock_intercept.assert_called_once_with(fake_channel, MockInterceptor.return_value)
        assert result is intercepted_channel

    def test_insecure_channel_keyword_arg_parses_host_port(self):
        """insecure_channel called with target= kwarg."""
        instrumentation = _make_client_instrumentation()
        fake_channel = MagicMock()
        wrapped = MagicMock(return_value=fake_channel)

        with patch("grpc.intercept_channel") as mock_intercept, patch(
            "elasticapm.contrib.grpc.client_interceptor._ClientInterceptor"
        ) as MockInterceptor:
            instrumentation.call(
                module="grpc",
                method="insecure_channel",
                wrapped=wrapped,
                instance=None,
                args=(),
                kwargs={"target": "myhost:8080"},
            )

        MockInterceptor.assert_called_once_with("myhost", 8080, secure=False)

    def test_secure_channel_sets_secure_flag(self):
        """secure_channel passes secure=True to _ClientInterceptor."""
        instrumentation = _make_client_instrumentation()
        wrapped = MagicMock(return_value=MagicMock())

        with patch("grpc.intercept_channel"), patch(
            "elasticapm.contrib.grpc.client_interceptor._ClientInterceptor"
        ) as MockInterceptor:
            instrumentation.call(
                module="grpc",
                method="secure_channel",
                wrapped=wrapped,
                instance=None,
                args=("myhost:443",),
                kwargs={},
            )

        MockInterceptor.assert_called_once_with("myhost", 443, secure=True)

    def test_host_without_port(self):
        """Target with no colon produces port=None."""
        instrumentation = _make_client_instrumentation()
        wrapped = MagicMock(return_value=MagicMock())

        with patch("grpc.intercept_channel"), patch(
            "elasticapm.contrib.grpc.client_interceptor._ClientInterceptor"
        ) as MockInterceptor:
            instrumentation.call(
                module="grpc",
                method="insecure_channel",
                wrapped=wrapped,
                instance=None,
                args=("myhost",),
                kwargs={},
            )

        MockInterceptor.assert_called_once_with("myhost", None, secure=False)

    def test_non_integer_port_becomes_none(self):
        """If port segment is not a valid integer, port is set to None."""
        instrumentation = _make_client_instrumentation()
        wrapped = MagicMock(return_value=MagicMock())

        with patch("grpc.intercept_channel"), patch(
            "elasticapm.contrib.grpc.client_interceptor._ClientInterceptor"
        ) as MockInterceptor:
            instrumentation.call(
                module="grpc",
                method="insecure_channel",
                wrapped=wrapped,
                instance=None,
                args=("myhost:notaport",),
                kwargs={},
            )

        MockInterceptor.assert_called_once_with("myhost", None, secure=False)

    def test_returns_intercepted_channel(self):
        """The return value is whatever grpc.intercept_channel returns."""
        instrumentation = _make_client_instrumentation()
        sentinel = MagicMock()
        wrapped = MagicMock(return_value=MagicMock())

        with patch("grpc.intercept_channel", return_value=sentinel), patch(
            "elasticapm.contrib.grpc.client_interceptor._ClientInterceptor"
        ):
            result = instrumentation.call(
                module="grpc",
                method="insecure_channel",
                wrapped=wrapped,
                instance=None,
                args=("host:1234",),
                kwargs={},
            )

        assert result is sentinel


# ---------------------------------------------------------------------------
# GRPCServerInstrumentation
# ---------------------------------------------------------------------------


class TestGRPCServerInstrumentation:
    def test_no_interceptors_adds_server_interceptor_via_kwargs(self):
        """With no existing interceptors, _ServerInterceptor is added via kwargs."""
        instrumentation = _make_server_instrumentation()
        fake_server = MagicMock()
        wrapped = MagicMock(return_value=fake_server)

        with patch("elasticapm.contrib.grpc.server_interceptor._ServerInterceptor") as MockInterceptor:
            result = instrumentation.call(
                module="grpc",
                method="server",
                wrapped=wrapped,
                instance=None,
                args=(),
                kwargs={},
            )

        interceptors_passed = wrapped.call_args[1]["interceptors"]
        assert interceptors_passed[0] is MockInterceptor.return_value
        assert result is fake_server

    def test_existing_interceptors_via_kwargs_prepends_server_interceptor(self):
        """_ServerInterceptor is inserted at index 0, before existing interceptors."""
        instrumentation = _make_server_instrumentation()
        existing = MagicMock()
        wrapped = MagicMock(return_value=MagicMock())

        with patch("elasticapm.contrib.grpc.server_interceptor._ServerInterceptor") as MockInterceptor:
            instrumentation.call(
                module="grpc",
                method="server",
                wrapped=wrapped,
                instance=None,
                args=(),
                kwargs={"interceptors": [existing]},
            )

        interceptors_passed = wrapped.call_args[1]["interceptors"]
        assert interceptors_passed[0] is MockInterceptor.return_value
        assert interceptors_passed[1] is existing

    def test_existing_interceptors_via_positional_args(self):
        """_ServerInterceptor is prepended when interceptors are in args[2]."""
        instrumentation = _make_server_instrumentation()
        existing = MagicMock()
        thread_pool = MagicMock()
        wrapped = MagicMock(return_value=MagicMock())

        # args[0]=thread_pool, args[1]=options, args[2]=interceptors
        with patch("elasticapm.contrib.grpc.server_interceptor._ServerInterceptor") as MockInterceptor:
            instrumentation.call(
                module="grpc",
                method="server",
                wrapped=wrapped,
                instance=None,
                args=(thread_pool, None, [existing]),
                kwargs={},
            )

        call_args = wrapped.call_args[0]
        interceptors_passed = call_args[2]
        assert interceptors_passed[0] is MockInterceptor.return_value
        assert interceptors_passed[1] is existing

    def test_no_interceptors_in_positional_args_uses_kwargs(self):
        """When args has fewer than 3 elements, interceptors go to kwargs."""
        instrumentation = _make_server_instrumentation()
        thread_pool = MagicMock()
        wrapped = MagicMock(return_value=MagicMock())

        with patch("elasticapm.contrib.grpc.server_interceptor._ServerInterceptor") as MockInterceptor:
            instrumentation.call(
                module="grpc",
                method="server",
                wrapped=wrapped,
                instance=None,
                args=(thread_pool,),
                kwargs={},
            )

        interceptors_passed = wrapped.call_args[1]["interceptors"]
        assert interceptors_passed[0] is MockInterceptor.return_value


# ---------------------------------------------------------------------------
# GRPCAsyncServerInstrumentation
# ---------------------------------------------------------------------------


class TestGRPCAsyncServerInstrumentation:
    def test_no_interceptors_adds_async_interceptor_via_kwargs(self):
        """With no existing interceptors, _AsyncServerInterceptor is added via kwargs."""
        instrumentation = _make_async_server_instrumentation()
        fake_server = MagicMock()
        wrapped = MagicMock(return_value=fake_server)

        with patch("elasticapm.contrib.grpc.async_server_interceptor._AsyncServerInterceptor") as MockInterceptor:
            result = instrumentation.call(
                module="grpc.aio",
                method="server",
                wrapped=wrapped,
                instance=None,
                args=(),
                kwargs={},
            )

        interceptors_passed = wrapped.call_args[1]["interceptors"]
        assert interceptors_passed[0] is MockInterceptor.return_value
        assert result is fake_server

    def test_existing_interceptors_via_kwargs_prepends_async_interceptor(self):
        """_AsyncServerInterceptor is inserted at index 0 before existing interceptors."""
        instrumentation = _make_async_server_instrumentation()
        existing = MagicMock()
        wrapped = MagicMock(return_value=MagicMock())

        with patch("elasticapm.contrib.grpc.async_server_interceptor._AsyncServerInterceptor") as MockInterceptor:
            instrumentation.call(
                module="grpc.aio",
                method="server",
                wrapped=wrapped,
                instance=None,
                args=(),
                kwargs={"interceptors": [existing]},
            )

        interceptors_passed = wrapped.call_args[1]["interceptors"]
        assert interceptors_passed[0] is MockInterceptor.return_value
        assert interceptors_passed[1] is existing

    def test_existing_interceptors_via_positional_args(self):
        """_AsyncServerInterceptor is prepended when interceptors are in args[2]."""
        instrumentation = _make_async_server_instrumentation()
        existing = MagicMock()
        thread_pool = MagicMock()
        wrapped = MagicMock(return_value=MagicMock())

        with patch("elasticapm.contrib.grpc.async_server_interceptor._AsyncServerInterceptor") as MockInterceptor:
            instrumentation.call(
                module="grpc.aio",
                method="server",
                wrapped=wrapped,
                instance=None,
                args=(thread_pool, None, [existing]),
                kwargs={},
            )

        call_args = wrapped.call_args[0]
        interceptors_passed = call_args[2]
        assert interceptors_passed[0] is MockInterceptor.return_value
        assert interceptors_passed[1] is existing

    def test_no_interceptors_in_short_positional_args_uses_kwargs(self):
        """When args has fewer than 3 elements, interceptors go to kwargs."""
        instrumentation = _make_async_server_instrumentation()
        thread_pool = MagicMock()
        wrapped = MagicMock(return_value=MagicMock())

        with patch("elasticapm.contrib.grpc.async_server_interceptor._AsyncServerInterceptor") as MockInterceptor:
            instrumentation.call(
                module="grpc.aio",
                method="server",
                wrapped=wrapped,
                instance=None,
                args=(thread_pool,),
                kwargs={},
            )

        interceptors_passed = wrapped.call_args[1]["interceptors"]
        assert interceptors_passed[0] is MockInterceptor.return_value



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import pytest  # isort:skip

pytest.importorskip("django")  # isort:skip

from os.path import join

import django
import pytest
from django.test.utils import override_settings

from elasticapm.conf.constants import TRANSACTION
from tests.contrib.django.conftest import BASE_TEMPLATE_DIR
from tests.utils.compat import middleware_setting

try:
    # Django 1.10+
    from django.urls import reverse
except ImportError:
    from django.core.urlresolvers import reverse

pytestmark = pytest.mark.django

# Testing Django 1.8+ backends
TEMPLATES = (
    {"BACKEND": "django.template.backends.django.DjangoTemplates", "DIRS": [BASE_TEMPLATE_DIR]},
    {"BACKEND": "django.template.backends.jinja2.Jinja2", "DIRS": [join(BASE_TEMPLATE_DIR, "jinja2")]},
)


def test_template_rendering(instrument, django_elasticapm_client, client):
    with override_settings(
        **middleware_setting(django.VERSION, ["elasticapm.contrib.django.middleware.TracingMiddleware"])
    ):
        client.get(reverse("render-heavy-template"))
        client.get(reverse("render-heavy-template"))
        client.get(reverse("render-heavy-template"))

    transactions = django_elasticapm_client.events[TRANSACTION]

    assert len(transactions) == 3
    spans = django_elasticapm_client.spans_for_transaction(transactions[0])
    assert len(spans) == 2, [t["name"] for t in spans]

    kinds = ["code", "template"]
    assert set([t["type"] for t in spans]) == set(kinds)

    assert spans[0]["type"] == "code"
    assert spans[0]["name"] == "something_expensive"
    assert spans[0]["parent_id"] == spans[1]["id"]

    assert spans[1]["type"] == "template"
    assert spans[1]["subtype"] == "django"
    assert spans[1]["action"] == "render"
    assert spans[1]["name"] == "list_users.html"
    assert spans[1]["parent_id"] == transactions[0]["id"]


@pytest.mark.skipif(django.VERSION < (1, 8), reason="Jinja2 support introduced with Django 1.8")
def test_template_rendering_django18_jinja2(instrument, django_elasticapm_client, client):
    with override_settings(
        TEMPLATES=TEMPLATES,
        **middleware_setting(django.VERSION, ["elasticapm.contrib.django.middleware.TracingMiddleware"])
    ):
        client.get(reverse("render-jinja2-template"))
        client.get(reverse("render-jinja2-template"))
        client.get(reverse("render-jinja2-template"))

    transactions = django_elasticapm_client.events[TRANSACTION]

    assert len(transactions) == 3
    spans = django_elasticapm_client.spans_for_transaction(transactions[0])
    assert len(spans) == 1, [t["name"] for t in spans]

    kinds = ["template"]
    assert set([t["type"] for t in spans]) == set(kinds)

    assert spans[0]["type"] == "template"
    assert spans[0]["subtype"] == "jinja2"
    assert spans[0]["action"] == "render"
    assert spans[0]["name"] == "jinja2_template.html"
    assert spans[0]["parent_id"] == transactions[0]["id"]



#  BSD 3-Clause License
#
#  Copyright (c) 2022, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import itertools
import os
import subprocess
import sys
import time
import urllib.parse
from collections import defaultdict

import pytest

from tests.fixtures import get_free_port, wait_for_open_port

grpc = pytest.importorskip("grpc")

pytestmark = pytest.mark.grpc


from .grpc_app.testgrpc_pb2 import Message
from .grpc_app.testgrpc_pb2_grpc import TestServiceStub


def setup_env(request, validating_httpserver):
    config = getattr(request, "param", {})
    env = {f"ELASTIC_APM_{k.upper()}": str(v) for k, v in config.items()}
    env.setdefault("ELASTIC_APM_SERVER_URL", validating_httpserver.url)
    return env


def setup_grpc_server(env):
    free_port = get_free_port()
    server_proc = subprocess.Popen(
        [os.path.join(sys.prefix, "bin", "python"), "-m", "tests.contrib.grpc.grpc_app.server", str(free_port)],
        stdout=sys.stdout,
        stderr=sys.stdout,
        env=env,
    )
    wait_for_open_port(free_port)
    return server_proc, free_port


@pytest.fixture()
def env_fixture(validating_httpserver, request):
    env = setup_env(request, validating_httpserver)
    return env


if hasattr(grpc, "aio"):
    grpc_server_fixture_params = ["async", "sync"]
else:
    grpc_server_fixture_params = ["sync"]


@pytest.fixture(params=grpc_server_fixture_params)
def grpc_client_and_server_url(env_fixture, request):
    env = {k: v for k, v in env_fixture.items()}
    if request.param == "async":
        env["GRPC_SERVER_ASYNC"] = "1"
    server_proc, free_port = setup_grpc_server(env)
    server_addr = f"localhost:{free_port}"
    test_channel = grpc.insecure_channel(server_addr)
    test_client = TestServiceStub(test_channel)
    yield test_client, server_addr
    server_proc.terminate()


def test_grpc_client_server_instrumentation(instrument, sending_elasticapm_client, grpc_client_and_server_url):
    grpc_client, grpc_server_url = grpc_client_and_server_url
    grpc_server_host, grpc_server_port = grpc_server_url.split(":")
    sending_elasticapm_client.begin_transaction("request")
    parsed_url = urllib.parse.urlparse(sending_elasticapm_client.httpserver.url)
    response = grpc_client.GetServerResponse(Message(message="foo"))
    sending_elasticapm_client.end_transaction("grpc-test")
    sending_elasticapm_client.close()
    payloads = sending_elasticapm_client.httpserver.payloads
    for i in range(1000):
        if len(sending_elasticapm_client.httpserver.payloads) > 2:
            break
        time.sleep(0.01)
    server_meta, server_transactions, server_spans, server_errors = extract_events_from_payload("grpc-server", payloads)
    client_meta, client_transactions, client_spans, client_errors = extract_events_from_payload("myapp", payloads)
    client_span = client_spans[0]
    assert len(server_errors) == 0

    assert client_span["trace_id"] == client_transactions[0]["trace_id"] == server_transactions[0]["trace_id"]

    assert server_transactions[0]["name"] == "/test.TestService/GetServerResponse"
    assert server_transactions[0]["outcome"] == "success"
    assert server_meta["service"]["framework"]["name"] == "grpc"
    assert server_meta["service"]["framework"]["version"] == grpc.__version__

    assert client_span["id"] == server_transactions[0]["parent_id"]
    assert client_span["name"] == "/test.TestService/GetServerResponse"
    assert client_span["type"] == "external"
    assert client_span["subtype"] == "grpc"
    assert client_span["context"]["destination"]["address"] == grpc_server_host
    assert client_span["context"]["destination"]["port"] == int(grpc_server_port)
    assert client_span["context"]["destination"]["service"]["resource"] == f"{grpc_server_host}:{grpc_server_port}"


def test_grpc_client_server_abort(instrument, sending_elasticapm_client, grpc_client_and_server_url):
    grpc_client, grpc_server_url = grpc_client_and_server_url
    grpc_server_host, grpc_server_port = grpc_server_url.split(":")
    sending_elasticapm_client.begin_transaction("request")
    parsed_url = urllib.parse.urlparse(sending_elasticapm_client.httpserver.url)
    with pytest.raises(Exception):
        response = grpc_client.GetServerResponseAbort(Message(message="foo"))
    sending_elasticapm_client.end_transaction("grpc-test")
    sending_elasticapm_client.close()
    payloads = sending_elasticapm_client.httpserver.payloads
    for i in range(1000):
        if len(sending_elasticapm_client.httpserver.payloads) > 3:
            break
        time.sleep(0.01)
    server_meta, server_transactions, server_spans, server_errors = extract_events_from_payload("grpc-server", payloads)
    client_meta, client_transactions, client_spans, client_errors = extract_events_from_payload("myapp", payloads)
    client_span = client_spans[0]
    assert len(server_errors) == 1

    assert client_span["trace_id"] == client_transactions[0]["trace_id"] == server_transactions[0]["trace_id"]

    assert server_transactions[0]["name"] == "/test.TestService/GetServerResponseAbort"
    assert server_transactions[0]["outcome"] == "failure"

    assert client_span["id"] == server_transactions[0]["parent_id"]
    assert client_span["name"] == "/test.TestService/GetServerResponseAbort"


def test_grpc_client_server_status(instrument, sending_elasticapm_client, grpc_client_and_server_url):
    grpc_client, grpc_server_url = grpc_client_and_server_url
    grpc_server_host, grpc_server_port = grpc_server_url.split(":")
    sending_elasticapm_client.begin_transaction("request")
    parsed_url = urllib.parse.urlparse(sending_elasticapm_client.httpserver.url)
    with pytest.raises(Exception):
        response = grpc_client.GetServerResponseUnavailable(Message(message="foo"))
    sending_elasticapm_client.end_transaction("grpc-test")
    sending_elasticapm_client.close()
    payloads = sending_elasticapm_client.httpserver.payloads
    for i in range(1000):
        if len(sending_elasticapm_client.httpserver.payloads) > 2:
            break
        time.sleep(0.01)
    server_meta, server_transactions, server_spans, server_errors = extract_events_from_payload("grpc-server", payloads)
    client_meta, client_transactions, client_spans, client_errors = extract_events_from_payload("myapp", payloads)
    client_span = client_spans[0]
    assert len(server_errors) == 0

    assert client_span["trace_id"] == client_transactions[0]["trace_id"] == server_transactions[0]["trace_id"]
    assert server_transactions[0]["outcome"] == "failure"
    assert client_span["id"] == server_transactions[0]["parent_id"]

    assert client_span["name"] == "/test.TestService/GetServerResponseUnavailable"


def test_grpc_client_server_exception(instrument, sending_elasticapm_client, grpc_client_and_server_url):
    grpc_client, grpc_server_url = grpc_client_and_server_url
    grpc_server_host, grpc_server_port = grpc_server_url.split(":")
    sending_elasticapm_client.begin_transaction("request")
    parsed_url = urllib.parse.urlparse(sending_elasticapm_client.httpserver.url)
    with pytest.raises(Exception):
        response = grpc_client.GetServerResponseException(Message(message="foo"))
    sending_elasticapm_client.end_transaction("grpc-test")
    sending_elasticapm_client.close()
    payloads = sending_elasticapm_client.httpserver.payloads
    for i in range(1000):
        if len(sending_elasticapm_client.httpserver.payloads) > 3:
            break
        time.sleep(0.01)
    server_meta, server_transactions, server_spans, server_errors = extract_events_from_payload("grpc-server", payloads)
    assert len(server_errors) == 1
    error = server_errors[0]
    assert error["transaction_id"] == server_transactions[0]["id"]


def test_grpc_client_unsampled_span(instrument, sending_elasticapm_client, grpc_client_and_server_url):
    grpc_client, grpc_server_url = grpc_client_and_server_url
    grpc_server_host, grpc_server_port = grpc_server_url.split(":")
    transaction = sending_elasticapm_client.begin_transaction("request")
    transaction.is_sampled = False
    parsed_url = urllib.parse.urlparse(sending_elasticapm_client.httpserver.url)
    response = grpc_client.GetServerResponse(Message(message="foo"))
    transaction.is_sampled = True
    sending_elasticapm_client.end_transaction("grpc-test")
    payloads = sending_elasticapm_client.httpserver.payloads
    for i in range(1000):
        if len(sending_elasticapm_client.httpserver.payloads) > 1:
            break
        time.sleep(0.01)
    transaction_data = payloads[1][1]["transaction"]
    assert transaction_data["span_count"]["started"] == 0


@pytest.mark.parametrize(
    "env_fixture",
    [
        {
            "recording": "False",
        }
    ],
    indirect=True,
)
def test_grpc_client_unsampled_transaction(instrument, sending_elasticapm_client, grpc_client_and_server_url):
    grpc_client, grpc_server_url = grpc_client_and_server_url
    grpc_client.GetServerResponse(Message(message="foo"))
    payloads = sending_elasticapm_client.httpserver.payloads
    for i in range(100):
        if len(sending_elasticapm_client.httpserver.payloads) > 1:
            break
        time.sleep(0.01)
    assert len(payloads) == 1  # only the server_version request


@pytest.mark.parametrize("elasticapm_client", [{"transaction_max_spans": 1}], indirect=True)
def test_grpc_client_max_spans(instrument, elasticapm_client, grpc_client_and_server_url):
    grpc_client, _ = grpc_client_and_server_url
    transaction = elasticapm_client.begin_transaction("request")
    _ = grpc_client.GetServerResponse(Message(message="foo"))
    _ = grpc_client.GetServerResponse(Message(message="bar"))
    elasticapm_client.end_transaction("grpc-test")
    assert transaction.dropped_spans == 1


def extract_events_from_payload(service_name, payloads):
    payloads = [
        payload for payload in payloads if payload and payload[0]["metadata"]["service"]["name"] == service_name
    ]
    events = defaultdict(list)
    for event in itertools.chain(*payloads):
        for k, v in event.items():
            events[k].append(v)
    return events["metadata"][0], events["transaction"], events["span"], events["error"]



#  BSD 3-Clause License
#
#  Copyright (c) 2025, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os

import pytest
import pytest_asyncio

from elasticapm.conf import constants

psycopg = pytest.importorskip("psycopg")  # isort:skip
pytestmark = [pytest.mark.psycopg, pytest.mark.asyncio]

if "POSTGRES_DB" not in os.environ:
    pytestmark.append(pytest.mark.skip("Skipping psycopg tests, no POSTGRES_DB environment variable set"))


def connect_kwargs():
    return {
        "dbname": os.environ.get("POSTGRES_DB", "elasticapm_test"),
        "user": os.environ.get("POSTGRES_USER", "postgres"),
        "password": os.environ.get("POSTGRES_PASSWORD", "postgres"),
        "host": os.environ.get("POSTGRES_HOST", None),
        "port": os.environ.get("POSTGRES_PORT", None),
    }


@pytest_asyncio.fixture(scope="function")
async def postgres_connection(request):
    conn = await psycopg.AsyncConnection.connect(**connect_kwargs())
    cursor = conn.cursor()
    await cursor.execute(
        "CREATE TABLE test(id int, name VARCHAR(5) NOT NULL);"
        "INSERT INTO test VALUES (1, 'one'), (2, 'two'), (3, 'three');"
    )

    yield conn

    # cleanup
    await cursor.execute("ROLLBACK")


async def test_cursor_execute_signature(instrument, postgres_connection, elasticapm_client):
    cursor = postgres_connection.cursor()
    record = await cursor.execute(query="SELECT 1", params=None, prepare=None, binary=None)
    assert record


async def test_cursor_executemany_signature(instrument, postgres_connection, elasticapm_client):
    cursor = postgres_connection.cursor()
    res = await cursor.executemany(
        query="INSERT INTO test VALUES (%s, %s)",
        params_seq=((4, "four"),),
        returning=False,
    )
    assert res is None


async def test_execute_with_sleep(instrument, postgres_connection, elasticapm_client):
    elasticapm_client.begin_transaction("test")
    cursor = postgres_connection.cursor()
    await cursor.execute("SELECT pg_sleep(0.1);")
    elasticapm_client.end_transaction("test", "OK")

    transaction = elasticapm_client.events[constants.TRANSACTION][0]
    spans = elasticapm_client.spans_for_transaction(transaction)

    assert len(spans) == 1
    span = spans[0]
    assert 100 < span["duration"] < 110
    assert transaction["id"] == span["transaction_id"]
    assert span["type"] == "db"
    assert span["subtype"] == "postgresql"
    assert span["action"] == "query"
    assert span["sync"] == False
    assert span["name"] == "SELECT FROM"


async def test_executemany(instrument, postgres_connection, elasticapm_client):
    elasticapm_client.begin_transaction("test")
    cursor = postgres_connection.cursor()
    await cursor.executemany("INSERT INTO test VALUES (%s, %s);", [(1, "uno"), (2, "due")])
    elasticapm_client.end_transaction("test", "OK")

    transaction = elasticapm_client.events[constants.TRANSACTION][0]
    spans = elasticapm_client.spans_for_transaction(transaction)

    assert len(spans) == 1
    span = spans[0]
    assert transaction["id"] == span["transaction_id"]
    assert span["subtype"] == "postgresql"
    assert span["action"] == "query"
    assert span["sync"] == False
    assert span["name"] == "INSERT INTO test"


async def test_server_cursor_execute(instrument, postgres_connection, elasticapm_client):
    cursor = postgres_connection.cursor(name="async_server_cursor")
    assert isinstance(cursor, psycopg.AsyncServerCursor)
    record = await cursor.execute(query="SELECT 1", params=None, binary=None)
    assert record



# -*- coding: utf-8 -*-

#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import absolute_import

import decimal
import uuid

from elasticapm.utils.encoding import enforce_label_format, shorten, transform


def test_transform_incorrect_unicode():
    x = "רונית מגן"

    result = transform(x)
    assert type(result) == str
    assert result == "רונית מגן"


def test_transform_correct_unicode():
    x = "רונית מגן"

    result = transform(x)
    assert type(result) == str
    assert result == x


def test_transform_bad_string():
    x = "The following character causes problems: \xd4".encode("latin-1")

    result = transform(x)
    assert type(result) == bytes
    assert result == "(Error decoding value)".encode("latin-1")


def test_transform_float():
    result = transform(13.0)
    assert type(result) == float
    assert result == 13.0


def test_transform_bool():
    result = transform(True)
    assert type(result) == bool
    assert result == True


def test_transform_int_subclass():
    class X(int):
        pass

    result = transform(X())
    assert type(result) == int
    assert result == 0


# def test_transform_bad_string():
#     x = 'The following character causes problems: \xd4'

#     result = transform(x)
#     assert result == '(Error decoding value)'

# def test_transform_model_instance():
#     instance = DuplicateKeyModel(foo='foo')

#     result = transform(instance)
#     assert result == '<DuplicateKeyModel: foo>'

# def test_transform_handles_gettext_lazy():
#     from django.utils.functional import lazy
#     def fake_gettext(to_translate):
#         return u'Igpay Atinlay'

#     fake_gettext_lazy = lazy(fake_gettext, str)

#     self.assertEquals(
#         pickle.loads(pickle.dumps(
#                 transform(fake_gettext_lazy("something")))),
#         u'Igpay Atinlay')


def test_transform_dict_keys():
    x = {"foo": "bar"}

    result = transform(x)
    assert type(result) == dict
    keys = list(result.keys())
    assert len(keys) == 1
    assert type(keys[0]), str
    assert keys[0] == "foo"


def test_transform_dict_keys_utf8_as_str():
    x = {"רונית מגן": "bar"}

    result = transform(x)
    assert type(result) == dict
    keys = list(result.keys())
    assert len(keys) == 1
    assert type(keys[0]), bytes
    assert keys[0] == "רונית מגן"


def test_transform_dict_keys_utf8_as_unicode():
    x = {str("\u05e8\u05d5\u05e0\u05d9\u05ea \u05de\u05d2\u05df"): "bar"}

    result = transform(x)
    keys = list(result.keys())
    assert len(keys) == 1
    assert type(keys[0]), str
    assert keys[0] == "\u05e8\u05d5\u05e0\u05d9\u05ea \u05de\u05d2\u05df"


def test_transform_uuid():
    x = uuid.uuid4()
    result = transform(x)
    assert result == repr(x)
    assert type(result), str


def test_transform_recursive():
    x = []
    x.append(x)

    result = transform(x)
    assert result == ["<...>"]


def test_transform_custom_repr():
    class Foo(object):
        def __elasticapm__(self):
            return "example"

    x = Foo()

    result = transform(x)
    assert result == "example"


def test_transform_broken_repr():
    class Foo(object):
        def __repr__(self):
            raise ValueError

    x = Foo()

    result = transform(x)
    expected = "<BadRepr: <class 'tests.utils.encoding.tests.test_transform_broken_repr.<locals>.Foo'>>"
    assert result == expected


def test_shorten_string():
    result = shorten("hello world!", string_length=5)
    assert len(result) == 5
    assert result == "he..."


def test_shorten_lists():
    result = shorten(list(range(500)), list_length=50)
    assert len(result) == 52
    assert result[-2] == "..."
    assert result[-1] == "(450 more elements)"


def test_shorten_sets():
    result = shorten(set(range(500)), list_length=50)
    assert len(result) == 52
    assert result[-2] == "..."
    assert result[-1] == "(450 more elements)"


def test_shorten_frozenset():
    result = shorten(frozenset(range(500)), list_length=50)
    assert len(result) == 52
    assert result[-2] == "..."
    assert result[-1] == "(450 more elements)"


def test_shorten_tuple():
    result = shorten(tuple(range(500)), list_length=50)
    assert len(result) == 52
    assert result[-2] == "..."
    assert result[-1] == "(450 more elements)"

    # def test_shorten_generator():
    #     result = shorten(xrange(500))
    #     assert len(result) == 52
    #     assert result[-2] == '...'
    #     assert result[-1] == '(450 more elements)'


def test_shorten_dict():
    result = shorten({k: k for k in range(500)}, dict_length=50)
    assert len(result) == 51
    assert result["<truncated>"] == "(450 more elements)"


def test_enforce_label_format():
    class MyObj(object):
        def __str__(self):
            return "OK"

        def __unicode__(self):
            return "OK"

    labels = enforce_label_format(
        {'a.b*c"d': MyObj(), "x": "x" * 1025, "int": 1, "float": 1.1, "decimal": decimal.Decimal("1")}
    )
    assert labels == {"a_b_c_d": "OK", "x": "x" * 1023 + "…", "int": 1, "float": 1.1, "decimal": decimal.Decimal("1")}



#  BSD 3-Clause License
#
#  Copyright (c) 2012, the Sentry Team, see AUTHORS for more details
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE


import random
import sys

from elasticapm.conf.constants import EXCEPTION_CHAIN_MAX_DEPTH
from elasticapm.utils import encoding, varmap
from elasticapm.utils.encoding import keyword_field, shorten, to_unicode
from elasticapm.utils.logging import get_logger
from elasticapm.utils.stacks import get_culprit, get_stack_info, iter_traceback_frames

__all__ = ("BaseEvent", "Exception", "Message")

logger = get_logger("elasticapm.events")


class BaseEvent(object):
    @staticmethod
    def to_string(client, data):
        raise NotImplementedError

    @staticmethod
    def capture(client, **kwargs):
        return {}


class Exception(BaseEvent):
    """
    Exceptions store the following metadata:

    - value: 'My exception value'
    - type: 'ClassName'
    - module '__builtin__' (i.e. __builtin__.TypeError)
    - frames: a list of serialized frames (see _get_traceback_frames)
    """

    @staticmethod
    def to_string(client, data):
        exc = data["exception"]
        if exc["value"]:
            return "%s: %s" % (exc["type"], exc["value"])
        return exc["type"]

    @staticmethod
    def get_hash(data):
        exc = data["exception"]
        output = [exc["type"]]
        for frame in data["stacktrace"]["frames"]:
            output.append(frame["module"])
            output.append(frame["function"])
        return output

    @staticmethod
    def capture(client, exc_info=None, **kwargs):
        culprit = exc_value = exc_type = exc_module = frames = exc_traceback = None
        new_exc_info = False
        if not exc_info or exc_info is True:
            new_exc_info = True
            exc_info = sys.exc_info()

        if exc_info == (None, None, None):
            raise ValueError("No exception found: capture_exception requires an active exception.")

        try:
            exc_type, exc_value, exc_traceback = exc_info

            frames = get_stack_info(
                iter_traceback_frames(exc_traceback, config=client.config),
                with_locals=client.config.collect_local_variables in ("errors", "all"),
                library_frame_context_lines=client.config.source_lines_error_library_frames,
                in_app_frame_context_lines=client.config.source_lines_error_app_frames,
                include_paths_re=client.include_paths_re,
                exclude_paths_re=client.exclude_paths_re,
                locals_processor_func=lambda local_var: varmap(
                    lambda k, val: shorten(
                        val,
                        list_length=client.config.local_var_list_max_length,
                        string_length=client.config.local_var_max_length,
                        dict_length=client.config.local_var_dict_max_length,
                    ),
                    local_var,
                ),
            )

            culprit = kwargs.get("culprit", None) or get_culprit(
                frames, client.config.include_paths, client.config.exclude_paths
            )

            if hasattr(exc_type, "__module__"):
                exc_module = exc_type.__module__
                exc_type = exc_type.__name__
            else:
                exc_module = None
                exc_type = exc_type.__name__
        finally:
            if new_exc_info:
                try:
                    del exc_info
                    del exc_traceback
                except Exception as e:
                    logger.exception(e)
        if "message" in kwargs:
            message = kwargs["message"]
        else:
            message = "%s: %s" % (exc_type, to_unicode(exc_value)) if exc_value else str(exc_type)

        message = encoding.long_field(message)

        data = {
            "id": "%032x" % random.getrandbits(128),
            "culprit": keyword_field(culprit),
            "exception": {
                "message": message,
                "type": keyword_field(str(exc_type)),
                "module": keyword_field(str(exc_module)),
                "stacktrace": frames,
            },
        }
        if hasattr(exc_value, "_elastic_apm_span_id"):
            data["parent_id"] = exc_value._elastic_apm_span_id
            del exc_value._elastic_apm_span_id
        depth = kwargs.get("_exc_chain_depth", 0)
        if depth > EXCEPTION_CHAIN_MAX_DEPTH:
            return
        cause = exc_value.__cause__
        chained_context = exc_value.__context__

        # we follow the pattern of Python itself here and only capture the chained exception
        # if cause is not None and __suppress_context__ is False
        if chained_context and not (exc_value.__suppress_context__ and cause is None):
            if cause:
                chained_exc_type = type(cause)
                chained_exc_value = cause
            else:
                chained_exc_type = type(chained_context)
                chained_exc_value = chained_context
            chained_exc_info = chained_exc_type, chained_exc_value, chained_context.__traceback__

            chained_cause = Exception.capture(
                client, exc_info=chained_exc_info, culprit="None", _exc_chain_depth=depth + 1
            )
            if chained_cause:
                data["exception"]["cause"] = [chained_cause["exception"]]
        return data


class Message(BaseEvent):
    """
    Messages store the following metadata:

    - message: 'My message from %s about %s'
    - params: ('foo', 'bar')
    """

    @staticmethod
    def to_string(client, data):
        return data["log"]["message"]

    @staticmethod
    def get_hash(data):
        msg = data["param_message"]
        return [msg["message"]]

    @staticmethod
    def capture(client, param_message=None, message=None, level=None, logger_name=None, **kwargs):
        if message:
            param_message = {"message": message}
        params = param_message.get("params")
        message = param_message["message"] % params if params else param_message["message"]
        data = kwargs.get("data", {})
        message_data = {
            "id": "%032x" % random.getrandbits(128),
            "log": {
                "level": keyword_field(level or "error"),
                "logger_name": keyword_field(logger_name or "__root__"),
                "message": message,
                "param_message": keyword_field(param_message["message"]),
            },
        }
        if isinstance(data.get("stacktrace"), dict):
            message_data["log"]["stacktrace"] = data["stacktrace"]["frames"]
        if kwargs.get("exception"):
            message_data["culprit"] = kwargs["exception"]["culprit"]
            message_data["exception"] = kwargs["exception"]["exception"]
        return message_data



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import os.path

from tests.utils.compat import middleware_setting

where_am_i = os.path.dirname(os.path.abspath(__file__))

BASE_TEMPLATE_DIR = os.path.join(where_am_i, "testapp", "templates")


def pytest_configure(config):
    try:
        from django.conf import settings
    except ImportError:
        settings = None
    if settings is not None and not settings.configured:
        import django

        settings_dict = dict(
            SECRET_KEY="42",
            DATABASES={
                "default": {
                    "ENGINE": "django.db.backends.sqlite3",
                    "NAME": "elasticapm_tests.db",
                    "TEST_NAME": "elasticapm_tests.db",
                    "TEST": {"NAME": "elasticapm_tests.db"},
                }
            },
            TEST_DATABASE_NAME="elasticapm_tests.db",
            INSTALLED_APPS=[
                "django.contrib.auth",
                "django.contrib.admin",
                "django.contrib.sessions",
                "django.contrib.sites",
                "django.contrib.redirects",
                "django.contrib.contenttypes",
                "elasticapm.contrib.django",
                "tests.contrib.django.testapp",
            ],
            ROOT_URLCONF="tests.contrib.django.testapp.urls",
            DEBUG=False,
            SITE_ID=1,
            BROKER_HOST="localhost",
            BROKER_PORT=5672,
            BROKER_USER="guest",
            BROKER_PASSWORD="guest",
            BROKER_VHOST="/",
            CELERY_ALWAYS_EAGER=True,
            TEMPLATE_DEBUG=False,
            TEMPLATE_DIRS=[BASE_TEMPLATE_DIR],
            ALLOWED_HOSTS=["*"],
            TEMPLATES=[
                {
                    "BACKEND": "django.template.backends.django.DjangoTemplates",
                    "DIRS": [BASE_TEMPLATE_DIR],
                    "OPTIONS": {
                        "context_processors": ["django.contrib.auth.context_processors.auth"],
                        "loaders": ["django.template.loaders.filesystem.Loader"],
                        "debug": False,
                    },
                }
            ],
            ELASTIC_APM={
                "METRICS_INTERVAL": "0ms",
                "TRANSPORT_CLASS": "tests.fixtures.DummyTransport",
                "SERVICE_NAME": "testapp",
                "CENTRAL_CONFIG": False,
                "CLOUD_PROVIDER": False,
            },  # avoid autostarting the metrics collector thread
        )
        settings_dict.update(
            **middleware_setting(
                django.VERSION,
                [
                    "django.contrib.sessions.middleware.SessionMiddleware",
                    "django.contrib.auth.middleware.AuthenticationMiddleware",
                    "django.contrib.messages.middleware.MessageMiddleware",
                ],
            )
        )
        settings.configure(**settings_dict)
        if hasattr(django, "setup"):
            django.setup()
